package ca.uwaterloo.crysp.attackSS;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;


import android.util.Log;

public class KNNClassifier {
	int k, numFeatures;
	ArrayList<Data> ds;
	String user;
	public KNNModel model;
	public boolean scaleFeatures = true;
	
	public KNNClassifier(int _k, int _numFeatures) {
		k = _k;
		ds = MainActivity.ds;
		numFeatures = _numFeatures;
		model = new KNNModel(_k, _numFeatures);
		
	}
	
	public void train(List<double[]> data, List<Integer> labels) {
		if (data == null || data.size() < 1) 
			throw new IllegalArgumentException("Provided data is invalid");
		model.data = data;
		model.labels = labels;
		if (scaleFeatures) {
			for (int i = 0; i < model.numFeatures; i++)
				model.scalingFactor[i] = 0;
				//scalingFactor[i] = Math.abs(data.get(0)[i]);
			for (int j = 0; j < data.size(); j++){//(double[] fv : data) {
				if (labels.get(j) ==  -1)
					continue;
				for (int i = 0; i < model.numFeatures; i++)
					if (model.scalingFactor[i] < Math.abs(data.get(j)[i]))
						model.scalingFactor[i] = Math.abs(data.get(j)[i]);
			}
			for (int i = 0; i < model.numFeatures; i++)
				if (model.scalingFactor[i] == 0)
					model.scalingFactor[i] = 1;
		}
		updateVictimScore();
	}
	
	public int classify(double [] featureVector) 
			throws IllegalStateException{
		List<ComputedDistance> computedDistance = 
				new ArrayList<ComputedDistance>();
		
		
		for (int j = 0; j < model.data.size(); j++){//(double[] fv : model.data) {
			double eucDistance = 0;
			for (int i = 0; i < model.numFeatures; i++)
				eucDistance += Math.pow(model.data.get(j)[i]/model.scalingFactor[i]
						- featureVector[i]/model.scalingFactor[i], 2);
			eucDistance = Math.sqrt(eucDistance);
			computedDistance.add(new ComputedDistance(model.labels.get(j), 
					eucDistance));
		}
		
		List<ComputedDistance> sortedDistance = sortDistances(computedDistance,
				model.k);
		return getMajorityLabel(sortedDistance);
	}
	
	public ArrayList<ComputedDistance> sortDistances(List<ComputedDistance> 
	computedDistance, int K) {
		ArrayList<ComputedDistance> sortedDistance = 
				new ArrayList<ComputedDistance>();
		for (int i = 0; i < K; i++) {
			double minDistance = Double.MAX_VALUE;
			int minIndex = -1;
			for (int j = 0; j < computedDistance.size(); j++) {
				if (minDistance > computedDistance.get(j).distance) {
					minIndex = j;
					minDistance = computedDistance.get(j).distance;
				}
			}
			if (minIndex == -1)
				break;
			sortedDistance.add(new ComputedDistance(
					computedDistance.get(minIndex).label, 
					computedDistance.get(minIndex).distance));
			computedDistance.remove(minIndex);
		}
		return sortedDistance;
	}

	public int getMajorityLabel(List<ComputedDistance> sortedDistance ) {
		if (sortedDistance.size() == 0 )
			return 0;
		Map <Integer, Integer> majoritySum = new HashMap<Integer, Integer>();
		for (ComputedDistance cd : sortedDistance) {
			if (majoritySum.containsKey(cd.label))
				majoritySum.put(cd.label, majoritySum.get(cd.label)+1);
			else
				majoritySum.put(cd.label, 1);
		}
		int majorityClassLabel = -1;
		int maxOccurences = -1;
		for (int classLabel : majoritySum.keySet()) {
			if (majoritySum.get(classLabel) > maxOccurences) {//XXX: break ties
				majorityClassLabel = classLabel;
				maxOccurences = majoritySum.get(classLabel);
			}
		}
		return majorityClassLabel;
	}
	
	public void updateVictimScore() {
		int sampleSize = MainActivity.positive.size()/2;
		ArrayList <double[]> positive = MainActivity.positive;
		ArrayList<Integer> victimScore = new ArrayList<Integer>();
		
		ArrayList<double []> negInstances = new ArrayList<double []>();
		for (int i = 0; i < MainActivity.trainingSet.size(); i++) {
			if (MainActivity.trainingLabels.get(i) == 1)
				continue;
			negInstances.add(MainActivity.trainingSet.get(i));
		}
		victimScore.addAll(classify(positive.subList(sampleSize, positive.size()),
			negInstances, positive.subList(0, sampleSize), true));
				
		victimScore.addAll(classify(positive.subList(0, sampleSize), negInstances, 
					positive.subList(sampleSize, positive.size()), true));

		int maxIndex = 0, maxScore = -1;
		for (int i = 0; i < victimScore.size(); i++)
			if (victimScore.get(i) > maxScore) {
				maxScore = victimScore.get(i);
				maxIndex = i;
			}
		MainActivity.targetIndex = maxIndex;
		MainActivity.populateRaw();
		//Log.e(maxIndex+"",maxScore+"");
		//for (int v : ds.get(user).selfScore)
		//	Log.e("score", v+"");
	}
	
	
	public ArrayList<Integer> classify(List<double[]> p, List <double[]> n, 
			List <double[]> u, boolean returnNeighborScore) {
		List<ComputedDistance> computedDistance = 
				new ArrayList<ComputedDistance>();
		
		ArrayList <Integer> scores = new ArrayList<Integer>();
		double [] _normVector = new double[numFeatures];
		for (int i = 0; i < numFeatures; i++) 
			_normVector[i] = 0;
		for (int i = 0; i < p.size(); i++) 
			for (int j = 0; j < numFeatures; j++) {
				double v = Math.abs(p.get(i)[j]);
				if (v >= _normVector[j]) 
					_normVector[j] = v;
			}		
		for (int j = 0; j < numFeatures; j++)
			if (_normVector[j] == 0)
				_normVector[j] = 1;
			//_normVector[j] = 3 / _normVector[j];
		
		for (double [] d : u) {
			computedDistance.clear();
			for (double [] e : p) {
				double dist = 0;
				for (int i = 1; i < numFeatures; i++) {
					//Log.e(i+"", dist+"");//index 1 is 0
					if (Double.isNaN(_normVector[i]))
						continue;
					dist += Math.pow((e[i]*_normVector[i])
							- (d[i]*_normVector[i]), 2);
				}
				dist = Math.sqrt(dist);
				computedDistance.add(new ComputedDistance(1, dist));
			}
			for (double [] e : n) {
				double dist = 0;
				for (int i = 1; i < numFeatures; i++) {
					if (Double.isNaN(_normVector[i]))
						continue;
					dist += Math.pow((e[i]*_normVector[i])
							- (d[i]*_normVector[i]), 2);
				}
				dist = Math.sqrt(dist);
				computedDistance.add(new ComputedDistance(-1, dist));
			}

			if(returnNeighborScore) {
				List<ComputedDistance> sortedDistance = 
						sortDistances(computedDistance, 25);//XXX
				scores.add(getNeighborScore(sortedDistance));
			}
			else {
				List<ComputedDistance> sortedDistance = 
						sortDistances(computedDistance, k);
				scores.add(getMajorityLabel(sortedDistance));
			}
			//	for (ComputedDistance cd : computedDistance)
			//		Log.e("cd", cd.label +":" +cd.distance);

		}
		return scores;
	}
	
	public int getNeighborScore(List<ComputedDistance> sortedDistance) {
		if (sortedDistance.size() == 0 )
			return 0;
		int neighborScore = 0;
		for (ComputedDistance cd : sortedDistance) 
			if (cd.label == 1)
				neighborScore++;
		return neighborScore;
	}
	
	public class ComputedDistance {
		int label;
		double distance;
		public ComputedDistance(int label, double distance) {
			this.label = label;
			this.distance = distance;
		}
	}

	public class KNNModel {
		/**
		 * A reference to training and testing instances
		 */
		public List<double[]> data;
		public List<Integer> labels;
		public double [] scalingFactor;
		/**
		 * Number of nearest neighbors 'k'
		 */
		public int k;
		
		/**
		 * Number of Features for the classifier
		 */
		public int numFeatures;
		
		/**
		 * constructor for KNNModel
		 * @param k Number of nearest neighbors 'k'
		 * @param numFeatures Number of Features for the classifier
		 */
		public KNNModel(int k, int numFeatures){
			this.k = k;
			this.numFeatures = numFeatures;
			this.scalingFactor = new double[numFeatures];
		}
	}
}
